﻿using System;
using System.Collections.Generic;
using System.Dynamic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Threading.Tasks;
using CsvHelper;

namespace eea_e1a_get
{
    public class Program
    {
        private static HttpClient Client = new HttpClient();
        public static async Task Main(string[] args)
        {
            //args: pollutant from_year to_year 
            //opt:  country city
            int firstYear = args.Length > 1 ? Convert.ToInt32(args[1]) : 2018;
            int lastYear = args.Length > 2 ? Convert.ToInt32(args[2]) : firstYear;
            string pollutant = args.Length > 0 ? args[0] : "PM10";
            string country = args.Length > 3 ? args[3] : "";
            string city = args.Length > 4 ? args[4] : "";

            for (int i = firstYear; i <= lastYear; i++)
            {
                Console.WriteLine("Processing year " + i);
                await Do(i, pollutant, country, city);
                Console.WriteLine("------------------------");
            }

        }

        private static async Task Do(int year, string pollutant, string country, string city)
        {
            try
            {
                string source = year < 2013 ? "Airbase" : "E1a";
                Dictionary<string, List<E1A>> data = new Dictionary<string, List<E1A>>();

                Console.WriteLine("Downloading metadata...");
                List<Metadata> metadata = await EEAService.GetMetadata();
                Console.WriteLine("Done");

                Console.WriteLine("Downloading datasource links...");
                var datasource_urls = EEAService.GetDatasources(country, city, source, year, GetPollutantId(pollutant));
                Console.WriteLine("Done");

                int count = 1;
                foreach (var url in datasource_urls)
                {
                    try
                    {
                        var records = await EEAService.GetE1A(url);
                        // One file can contain multiple resolutions
                        var resolutions = records.Select(p => p.AveragingTime).Distinct().ToList();
                        foreach (var resolution in resolutions)
                        {
                            if (!data.ContainsKey(resolution))
                                data.Add(resolution, new List<E1A>());

                            data[resolution].AddRange(records.Where(p => p.AveragingTime == resolution));
                        }


                        Console.WriteLine("Downloaded " + count + " of " + datasource_urls.Count() + " datasources");
                        count++;

                    }
                    catch (System.Exception e)
                    {
                        using (StreamWriter sw = File.AppendText(source + "-" + year + "-" + pollutant + "-Exception.txt"))
                        {
                            sw.WriteLine("Could not parse: " + url);
                            sw.WriteLine("Exception: " + e.Message);
                            sw.WriteLine("-----------------------------------------");
                        }
                        Console.WriteLine("Could not parse: " + url);
                    }
                }

                foreach (var item in data)
                {
                    Resolution resolution = new Resolution(item.Key);

                    Console.WriteLine("Processing " + resolution.ResolutionAsString);
                    Console.WriteLine("---");
                    Console.WriteLine("Grouping timeseries...");
                    int observationsInAYear = resolution.GetObservationsInAYear(year);

                    var timeseries =
                        (from d in item.Value
                         group d by new { d.SamplingPoint, d.AirQualityStation, d.UnitOfMeasurement } into g
                         select new Timeserie
                         {
                             SamplingPoint = g.Key.SamplingPoint,
                             AirQualityStation = g.Key.AirQualityStation,
                             UnitOfMeasurement = g.Key.UnitOfMeasurement,
                             Resolution = resolution,
                             Coverage = Math.Round(((g.Count() / Convert.ToDouble(observationsInAYear)) * 100)),
                             Values = g.ToList()
                         })
                        .Where(p => p.Coverage >= 75 && metadata.Any(m => m.SamplingPoint == p.SamplingPoint))
                        .ToList();
                    Console.WriteLine("Done");

                    CSV csv = new CSV();

                    Console.WriteLine("Adding metadata...");

                    csv.AddCell("stationid", "samplingpoint", "stationid");
                    csv.AddCell("areatype", "samplingpoint", "areatype");
                    csv.AddCell("stationtype", "samplingpoint", "stationtype");
                    csv.AddCell("latitude", "samplingpoint", "latitude");
                    csv.AddCell("longitude", "samplingpoint", "longitude");
                    csv.AddCell("unit", "samplingpoint", "unit");

                    foreach (var t in timeseries)
                    {
                        var m = metadata.Where(p => p.SamplingPoint == t.SamplingPoint).FirstOrDefault();
                        if (m != null)
                        {
                            csv.AddCell("stationid", t.SamplingPoint, m.AirQualityStation);
                            csv.AddCell("stationtype", t.SamplingPoint, m.AirQualityStationType);
                            csv.AddCell("areatype", t.SamplingPoint, m.AirQualityStationArea);
                            csv.AddCell("latitude", t.SamplingPoint, m.Latitude);
                            csv.AddCell("longitude", t.SamplingPoint, m.Longitude);
                            csv.AddCell("unit", t.SamplingPoint, t.UnitOfMeasurement);
                        }
                        else
                        {
                            using (StreamWriter sw = File.AppendText(source + "-" + year + "-" + pollutant + "-Exception.txt"))
                            {
                                sw.WriteLine("Could not find metadata for " + t.SamplingPoint);
                                sw.WriteLine("-----------------------------------------");
                            }
                            Console.WriteLine("Could not find metadata for " + t.SamplingPoint);
                        }
                    }
                    Console.WriteLine("Done");

                    Console.WriteLine("Adding dates...");
                    var defaultValues = timeseries.Select(p => p.SamplingPoint).Prepend("samplingpoint").Distinct().ToDictionary(v => v, v => (object)-990);
                    DateTime dt = new DateTime(year, 1, 1, 0, 0, 0, DateTimeKind.Utc);
                    for (int i = 0; i < observationsInAYear; i++)
                    {
                        var current = resolution.AddTimeStep(dt, i);
                        var dateAsString = current.ToString("yyyy-MM-ddTHH:mm:ssZ");

                        defaultValues["samplingpoint"] = dateAsString;
                        Dictionary<string, object> cells = new Dictionary<string, object>(defaultValues);
                        csv.AddRow(current, cells);
                    }
                    Console.WriteLine("Done");


                    Console.WriteLine("Adding values...");
                    var timevalues = timeseries.SelectMany(p => p.Values).ToList();
                    foreach (var d in timevalues)
                    {
                        if (resolution.ToUniversalTime(d.DatetimeEnd) == new DateTime(2000, 03, 29))
                        {
                            string t = "";
                        }
                        csv.EditCell(resolution.ToUniversalTime(d.DatetimeEnd), d.SamplingPoint, d.Concentration);
                    }
                    Console.WriteLine("Done");


                    Console.WriteLine("Converting objects...");
                    var converted = csv.ToObject();
                    Console.WriteLine("Done");

                    Console.WriteLine("Writing to csv...");
                    using (var writer = new StreamWriter(source + "-" + year + "-" + pollutant + "-" + resolution.ResolutionAsString + ".csv"))
                    using (var c = new CsvWriter(writer, CultureInfo.InvariantCulture))
                    {
                        c.WriteRecords(converted);
                    }
                    Console.WriteLine("Done");
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Something unexpected happen");
                Console.WriteLine("Exception: " + e.Message);
            }
        }

        private static dynamic DictionaryToObject(Dictionary<string, object> dict)
        {
            IDictionary<string, object> eo = new ExpandoObject() as IDictionary<string, object>;
            foreach (KeyValuePair<string, object> kvp in dict)
            {
                eo.Add(kvp);
            }
            return eo;
        }

        private static int GetPollutantId(string pollutant)
        {
            if (pollutant == "SO2")
                return 1;
            if (pollutant == "PM2.5")
                return 6001;
            if (pollutant == "PM10")
                return 5;
            if (pollutant == "NO2")
                return 8;
            if (pollutant == "O3")
                return 7;
            if (pollutant == "NO")
                return 38;

            return -1;
        }
    }



}
